# vim: set fileencodings=utf-8
# -*- coding: utf-8; -*-
# PYTHON_ARGCOMPLETE_OK
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


from __future__ import print_function

import logging
import os
import re
import sys
import urllib3
from appdirs import user_data_dir, user_config_dir
from argparse import ArgumentParser
from collections import namedtuple
from crate.client import connect
from crate.client.exceptions import ConnectionError, ProgrammingError
from distutils.version import StrictVersion
from urllib3.exceptions import LocationParseError

from .commands import built_in_commands, Command
from .config import Configuration, ConfigurationError
from .outputs import OutputWriter
from .printer import ColorPrinter, PrintWrapper
from .sysinfo import SysInfoCommand
from ..crash import __version__ as crash_version

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


try:
    from logging import NullHandler
except ImportError:
    from logging import Handler

    class NullHandler(Handler):

        def emit(self, record):
            pass


logging.getLogger('crate').addHandler(NullHandler())

USER_DATA_DIR = user_data_dir("Crate", "Crate")
HISTORY_FILE_NAME = 'crash_history'
HISTORY_PATH = os.path.join(USER_DATA_DIR, HISTORY_FILE_NAME)

USER_CONFIG_DIR = user_config_dir("Crate", "Crate")
CONFIG_FILE_NAME = 'crash.cfg'
CONFIG_PATH = os.path.join(USER_CONFIG_DIR, CONFIG_FILE_NAME)

Result = namedtuple('Result', ['cols',
                               'rows',
                               'rowcount',
                               'duration',
                               'output_width'])

TABLE_SCHEMA_MIN_VERSION = StrictVersion("0.57.0")


def parse_config_path(args=sys.argv):
    """
    Preprocess sys.argv and extract --config argument.
    """

    config = CONFIG_PATH
    if '--config' in args:
        idx = args.index('--config')
        if len(args) > idx + 1:
            config = args.pop(idx + 1)
        _ = args.pop(idx)
    return config


def parse_args(parser):
    """
    Parse sys.argv arguments with given parser
    """
    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except ImportError:
        pass

    return parser.parse_args()


def boolean(v):
    if str(v).lower() in ("yes", "true", "t", "1"):
        return True
    elif str(v).lower() in ("no", "false", "f", "0"):
        return False
    else:
        raise ValueError('not a boolean value')


def get_parser(output_formats=[], conf=None):
    """
    Create an argument parser that reads default values from a
    configuration file if provided.
    """

    def _conf_or_default(key, value):
        return value if conf is None else conf.get_or_set(key, value)

    parser = ArgumentParser(description='crate shell')
    parser.add_argument('-v', '--verbose', action='count',
                        dest='verbose', default=_conf_or_default('verbosity', 0),
                        help='use -v to get debug output')
    parser.add_argument('-A', '--no-autocomplete', action='store_false',
                        dest='autocomplete',
                        default=_conf_or_default('autocomplete', True),
                        help='use -A to disable SQL autocompletion')
    parser.add_argument('-a', '--autocapitalize', action='store_true',
                        dest='autocapitalize',
                        default=False,
                        help='use -a to enable experimental auto-capitalization of SQL keywords')
    parser.add_argument('-U', '--username', type=str,
                        help='the username to authenticate in the database')
    parser.add_argument('--history', type=str,
                        help='the history file to use', default=HISTORY_PATH)
    parser.add_argument('--config', type=str,
                        help='the configuration file to use', default=CONFIG_PATH)

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-c', '--command', type=str,
                       help='execute sql statement')
    group.add_argument('--sysinfo', action='store_true', default=False,
                       help='show system information')

    parser.add_argument('--hosts', type=str, nargs='*',
                        default=_conf_or_default('hosts', ['localhost:4200']),
                        help='the crate hosts to connect to', metavar='HOST')
    parser.add_argument('--verify-ssl', type=boolean, default=True,
                        help='force verification of the SSL certificate of the server')
    parser.add_argument('--cert-file', type=file_with_permissions,
                        help='path to the client certificate file')
    parser.add_argument('--key-file', type=file_with_permissions,
                        help='path to the key file of the client certificate')
    parser.add_argument('--ca-cert-file', type=file_with_permissions,
                        help='path to the CA certificate file')
    parser.add_argument('--format', type=str,
                        default=_conf_or_default('format', 'tabular'),
                        choices=output_formats,
                        help='output format of the sql response', metavar='FORMAT')
    parser.add_argument('--version', action='store_true', default=False,
                        help='show crash version and exit')

    return parser


def noargs_command(fn):
    def inner_fn(self, *args):
        if len(args):
            self.logger.critical("Command does not take any arguments.")
            return
        return fn(self, *args)
    inner_fn.__doc__ = fn.__doc__
    return inner_fn


class CrateCmd(object):

    def __init__(self,
                 output_writer=None,
                 connection=None,
                 error_trace=False,
                 is_tty=True,
                 autocomplete=True,
                 autocapitalize=True,
                 verify_ssl=True,
                 cert_file=None,
                 key_file=None,
                 ca_cert_file=None,
                 username=None):
        self.error_trace = error_trace
        self.connection = connection or connect(error_trace=error_trace)
        self.cursor = self.connection.cursor()
        self.output_writer = output_writer or OutputWriter(
            PrintWrapper(), is_tty)
        self.lines = []
        self.exit_code = 0
        self.expanded_mode = False
        self.sys_info_cmd = SysInfoCommand(self)
        self.commands = {
            'q': self._quit,
            'c': self._connect,
            'connect': self._connect,
            'dt': self._show_tables,
            'sysinfo': self.sys_info_cmd.execute,
        }
        self.commands.update(built_in_commands)
        self.logger = ColorPrinter(is_tty)
        self._autocomplete = autocomplete
        self._autocapitalize = autocapitalize
        self.username = username
        self.verify_ssl = verify_ssl
        self.cert_file = cert_file
        self.key_file = key_file
        self.ca_cert_file = ca_cert_file

    def get_num_columns(self):
        return 80

    def should_autocomplete(self):
        return self._autocomplete

    def should_autocapitalize(self):
        return self._autocapitalize

    def pprint(self, rows, cols):
        result = Result(cols,
                        rows,
                        self.cursor.rowcount,
                        self.cursor.duration,
                        self.get_num_columns())
        self.output_writer.write(result)

    def process(self, text):
        if text.lstrip().startswith('--'):
            return
        text = text.rstrip()
        if text.startswith('\\') and not self.lines:
            self._try_exec_cmd(text.lstrip('\\'))
        elif text.endswith(';'):
            line = text.rstrip(';')
            if self.lines:
                self.lines.append(line)
                line = ' '.join(self.lines)
                self._exec(line)
                self.lines[:] = []
            else:
                self._exec(line)
        elif text:
            self.lines.append(text)

    def exit(self):
        if self.lines:
            self._exec(' '.join(self.lines))
            self.lines[:] = []

    @noargs_command
    def _show_tables(self, *args):
        """ print the existing tables within the 'doc' schema """
        schema_name = \
            "table_schema" if self.connection.lowest_server_version \
            >= TABLE_SCHEMA_MIN_VERSION else "schema_name"

        self._exec("select format('%s.%s', {schema}, table_name) as name "
                   "from information_schema.tables "
                   "where {schema} not in ('sys','information_schema', 'pg_catalog')"
                   .format(schema=schema_name))

    @noargs_command
    def _quit(self, *args):
        """ quit crash """
        self.logger.warn(u'Bye!')
        sys.exit(self.exit_code)

    def is_conn_available(self):
        if self.connection.lowest_server_version == StrictVersion("0.0.0"):
            return False
        else:
            return True

    def _do_connect(self):
        self.connection = connect(servers=self.last_connected_servers, error_trace=self.error_trace,
                                  verify_ssl_cert=self.verify_ssl, cert_file=self.cert_file, key_file=self.key_file,
                                  ca_cert=self.ca_cert_file, username=self.username)
        self.cursor = self.connection.cursor()

    def _connect(self, server):
        """ connect to the given server, e.g.: \connect localhost:4200 """
        self.last_connected_servers = server
        self._do_connect()
        results = []
        failed = 0
        client = self.connection.client
        for server in client.active_servers:
            try:
                infos = client.server_infos(server)
            except ConnectionError as e:
                failed += 1
                results.append([server, None, '0.0.0', False, e.message])
            else:
                results.append(infos + (True, 'OK', ))
        self.pprint(results,
                    ['server_url', 'node_name', 'version', 'connected', 'message'])
        if failed == len(results):
            self.logger.critical('CONNECT ERROR')
        else:
            # Execute cluster/node checks only in verbose mode
            if self.error_trace:
                self.logger.info('CONNECT OK')
                SysInfoCommand.CLUSTER_INFO['information_schema_query'] = \
                get_information_schema_query(self.connection.lowest_server_version)
                # check for failing node and cluster checks
                built_in_commands['check'](self, startup=True)

    def _try_exec_cmd(self, line):
        words = line.split(' ', 1)
        if not words or not words[0]:
            return False
        cmd = self.commands.get(words[0].lower().rstrip(';'))
        if len(words) > 1:
            words[1] = words[1].rstrip(';')
        if cmd:
            try:
                if isinstance(cmd, Command):
                    message = cmd(self, *words[1:])
                else:
                    message = cmd(*words[1:])
            except TypeError as e:
                self.logger.critical(getattr(e, 'message', None) or repr(e))
                doc = cmd.__doc__
                if doc and not doc.isspace():
                    self.logger.info('help: {0}'.format(words[0].lower()))
                    self.logger.info(cmd.__doc__)
            except Exception as e:
                self.logger.critical(getattr(e, 'message', None) or repr(e))
            else:
                if message:
                    self.logger.info(message)
            return True
        else:
            self.logger.critical(
                'Unknown command. Type \? for a full list of available commands.')
        return False

    def _exec(self, line):
        success = self.execute(line)
        self.exit_code = self.exit_code or int(not success)

    def _execute(self, statement):
        try:
            self.cursor.execute(statement)
            return True
        except ConnectionError as e:
            if self.error_trace:
                self.logger.warn(str(e))
            self.logger.warn(
                'Use \connect <server> to connect to one or more servers first.')
        except ProgrammingError as e:
            self.logger.critical(e.message)
            if self.error_trace and e.error_trace:
                self.logger.critical('\n' + e.error_trace)
        return False

    def execute(self, statement):
        success = self._execute(statement)
        if not success:
            return False
        cur = self.cursor
        duration = ''
        if cur.duration > -1:
            duration = ' ({0:.3f} sec)'.format(float(cur.duration) / 1000.0)
        print_vars = {
            'command': stmt_type(statement),
            'rowcount': cur.rowcount,
            's': 's'[cur.rowcount == 1:],
            'duration': duration
        }
        if cur.description:
            self.pprint(cur.fetchall(), [c[0] for c in cur.description])
            tmpl = '{command} {rowcount} row{s} in set{duration}'
        else:
            tmpl = '{command} OK, {rowcount} row{s} affected {duration}'
        self.logger.info(tmpl.format(**print_vars))
        return True


def stmt_type(statement):
    """
    Extract type of statement, e.g. SELECT, INSERT, UPDATE, DELETE, ...
    """
    return re.findall('[\w]+', statement)[0].upper()


def get_stdin():
    """
    Get data from stdin, if any
    """
    if not sys.stdin.isatty():
        for line in sys.stdin:
            yield line
    return


def host_and_port(host_or_port):
    """
    Return full hostname/IP + port, possible input formats are:
      * host:port  -> host:port
      * :          -> localhost:4200
      * :port      -> localhost:port
      * host       -> host:4200
    """
    if ':' in host_or_port:
        if len(host_or_port) == 1:
            return 'localhost:4200'
        elif host_or_port.startswith(':'):
            return 'localhost' + host_or_port
        return host_or_port
    return host_or_port + ':4200'

def get_information_schema_query(lowest_server_version):
    schema_name = \
        "table_schema" if lowest_server_version >= \
        TABLE_SCHEMA_MIN_VERSION else "schema_name"

    information_schema_query = \
        """ select count(distinct(table_name))
                as number_of_tables
            from information_schema.tables
            where {schema}
            not in ('information_schema', 'sys', 'pg_catalog') """

    return information_schema_query.format(schema=schema_name)

def main():
    is_tty = sys.stdout.isatty()
    printer = ColorPrinter(is_tty)
    output_writer = OutputWriter(PrintWrapper(), is_tty)

    config = parse_config_path()
    conf = None
    try:
        conf = Configuration(config)
    except ConfigurationError as e:
        printer.warn(str(e))
        parser = get_parser(output_writer.formats)
        parser.print_usage()
        sys.exit(1)
    parser = get_parser(output_writer.formats, conf=conf)
    try:
        args = parse_args(parser)
    except Exception as e:
        printer.warn(str(e))
        sys.exit(1)
    output_writer.output_format = args.format

    if args.version:
        printer.info(crash_version)
        sys.exit(0)

    crate_hosts = [host_and_port(h) for h in args.hosts]
    error_trace = args.verbose > 0
    try:
        cmd = _create_cmd(crate_hosts, error_trace, output_writer, is_tty, args)
    except (ProgrammingError, LocationParseError) as e:
        printer.warn(str(e))
        sys.exit(1)

    cmd._connect(crate_hosts)
    done = False
    stdin_data = get_stdin()
    if args.sysinfo:
        cmd.output_writer.output_format = 'mixed'
        cmd.sys_info_cmd.execute()
        done = True
    if args.command:
        cmd.process(args.command)
        done = True
    elif stdin_data:
        for data in stdin_data:
            cmd.process(data)
            done = True
    if not done:
        from .repl import loop
        loop(cmd, args.history)
    conf.save()
    cmd.exit()
    sys.exit(cmd.exit_code)

def _create_cmd(crate_hosts, error_trace, output_writer, is_tty, args, timeout=None):
    conn = connect(crate_hosts,
                   verify_ssl_cert=args.verify_ssl,
                   cert_file=args.cert_file,
                   key_file=args.key_file,
                   ca_cert=args.ca_cert_file,
                   username=args.username,
                   timeout=timeout)
    return CrateCmd(connection=conn,
                    error_trace=error_trace,
                    output_writer=output_writer,
                    is_tty=is_tty,
                    autocomplete=args.autocomplete,
                    autocapitalize=args.autocapitalize,
                    verify_ssl=args.verify_ssl,
                    cert_file=args.cert_file,
                    key_file=args.key_file,
                    ca_cert_file=args.ca_cert_file,
                    username=args.username)

def file_with_permissions(path):
    open(path, 'r').close()
    return path

if __name__ == '__main__':
    main()



# vim: set fileencodings=utf-8
# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import os
import re

from pygments.lexers.sql import SqlLexer
from pygments.style import Style
from pygments.token import (Keyword,
                            Comment,
                            Operator,
                            Number,
                            Literal,
                            String,
                            Error,
                            Token)

from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.history import FileHistory
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import Condition, IsDone, HasFocus
from prompt_toolkit import CommandLineInterface, AbortAction, Application
from prompt_toolkit.interface import AcceptAction
from prompt_toolkit.styles import PygmentsStyle
from prompt_toolkit.enums import DEFAULT_BUFFER, EditingMode
from prompt_toolkit.layout.processors import (
    HighlightMatchingBracketProcessor,
    ConditionalProcessor
)
from prompt_toolkit.key_binding.manager import KeyBindingManager
from prompt_toolkit.keys import Keys
from prompt_toolkit.shortcuts import (create_output,
                                      create_eventloop)

from .commands import Command
from .layout import create_layout

MAX_HISTORY_LENGTH = 10000


def _get_editing_mode():
    files = ['/etc/inputrc', os.path.expanduser('~/.inputrc')]
    for filepath in files:
        try:
            with open(filepath, 'r') as f:
                for line in f:
                    if line.strip() == 'set editing-mode vi':
                        return EditingMode.VI
        except IOError:
            continue
    return EditingMode.EMACS


class CrateStyle(Style):
    default_style = "noinherit"
    styles = {
        Keyword: 'bold #4b95a3',
        Comment: '#757265',
        Operator: '#e83131',
        Number: '#be61ff',
        Literal: '#ae81ff',
        String: '#f4a33d',
        Error: '#ff3300',
        Token.Toolbar.Status.Key: '#4b95a3',
    }


class TruncatedFileHistory(FileHistory):

    def __init__(self, filename, max_length=1000):
        super(TruncatedFileHistory, self).__init__(filename)
        base = os.path.dirname(filename)
        if not os.path.exists(base):
            os.makedirs(base)
        if not os.path.exists(filename):
            with open(filename, 'a'):
                os.utime(filename, None)
        self.max_length = max_length

    def append(self, string):
        self.strings = self.strings[:max(0, self.max_length - 1)]
        super(TruncatedFileHistory, self).append(string)


class SQLCompleter(Completer):
    keywords = [
        "abs", "absolute", "action", "add", "after", "alias", "all", "allocate",
        "alter", "always", "analyzer", "and", "any", "are", "array", "array_agg",
        "array_max_cardinality", "as", "asc", "asensitive", "assertion",
        "asymmetric", "at", "atomic", "authorization", "avg", "before", "begin",
        "begin_frame", "begin_partition", "between", "bigint", "binary", "bit",
        "bit_length", "blob", "boolean", "both", "breadth", "by", "byte", "call",
        "called", "cardinality", "cascade", "cascaded", "case", "cast",
        "catalog", "catalogs", "ceil", "ceiling", "char", "char_filters",
        "char_length", "character", "character_length", "check", "clob", "close",
        "clustered", "coalesce", "collate", "collation", "collect", "column",
        "columns", "commit", "condition", "connect", "connection", "constraint",
        "constraints", "constructor", "contains", "continue", "convert", "copy",
        "corr", "corresponding", "count", "covar_pop", "covar_samp", "create",
        "cross", "cube", "cume_dist", "current", "current_catalog",
        "current_date", "current_path", "current_role", "current_row",
        "current_schema", "current_time", "current_timestamp", "current_user",
        "cursor", "cycle", "data", "date", "day", "deallocate", "dec", "decimal",
        "declare", "default", "deferrable", "deferred", "delete", "deny", "dense_rank",
        "depth", "deref", "desc", "describe", "descriptor", "deterministic",
        "diagnostics", "directory", "disconnect", "distinct", "distributed",
        "do", "domain", "double", "drop", "duplicate", "dynamic", "each",
        "element", "else", "elseif", "end", "end_exec", "end_frame",
        "end_partition", "equals", "escape", "every", "except", "exception",
        "exec", "execute", "exists", "exit", "explain", "extends", "external",
        "extract", "false", "fetch", "filter", "first", "first_value", "float",
        "for", "foreign", "format", "found", "frame_row", "free", "from", "full",
        "fulltext", "function", "functions", "fusion", "general", "generated",
        "geo_point", "geo_shape", "get", "global", "go", "goto", "grant",
        "group", "grouping", "groups", "handler", "having", "hold", "hour",
        "identity", "if", "ignored", "immediate", "in", "index", "indicator",
        "initially", "inner", "inout", "input", "insensitive", "insert", "int",
        "integer", "intersect", "intersection", "interval", "into", "ip", "is",
        "isolation", "iterate", "join", "key", "kill", "language", "large",
        "last", "last_value", "lateral", "lead", "leading", "leave", "left",
        "level", "like", "like_regex", "limit", "ln", "local", "localtime",
        "localtimestamp", "locator", "long", "loop", "lower", "map", "match",
        "max", "member", "merge", "method", "min", "minute", "mod", "modifies",
        "module", "month", "multiset", "names", "national", "natural", "nchar",
        "nclob", "new", "next", "no", "none", "normalize", "not", "nth_value",
        "ntile", "null", "nullif", "nulls", "numeric", "object", "octet_length",
        "of", "off", "offset", "old", "on", "only", "open", "optimize", "option",
        "or", "order", "ordinality", "out", "outer", "output", "over",
        "overlaps", "overlay", "pad", "parameter", "partial", "partition",
        "partitioned", "partitions", "path", "percent", "percent_rank",
        "percentile_cont", "percentile_disc", "period", "persistent", "plain",
        "portion", "position", "position_regex", "power", "precedes",
        "preceding", "precision", "prepare", "preserve", "primary",
        "primary key", "prior", "privileges", "procedure", "public", "range",
        "rank", "read", "reads", "real", "recursive", "ref", "references",
        "referencing", "refresh", "regr_avgx", "regr_avgy", "regr_count",
        "regr_intercept", "regr_r2", "regr_slope", "regr_sxx",
        "regr_sxyregr_syy", "relative", "release", "rename", "repeat", "repository",
        "reset", "resignal", "restore", "restrict", "result", "return",
        "returns", "revoke", "right", "role", "rollback", "rollup", "routine",
        "row", "row_number", "rows", "savepoint", "schema", "schemas", "scope",
        "scroll", "search", "second", "section", "select", "sensitive",
        "session", "session_user", "set", "sets", "shards", "short", "show",
        "signal", "similar", "size", "smallint", "snapshot", "some", "space",
        "specific", "specifictype", "sql", "sqlcode", "sqlerror", "sqlexception",
        "sqlstate", "sqlwarning", "sqrt", "start", "state", "static",
        "stddev_pop", "stddev_samp", "stratify", "stratify", "strict", "string",
        "submultiset", "substring", "substring_regex", "succeedsblob", "sum",
        "symmetric", "system", "system_time", "system_user", "table", "tables",
        "tablesample", "temporary", "text", "then", "time", "timestamp",
        "timezone_hour", "timezone_minute", "to", "token_filters", "tokenizer",
        "trailing", "transaction", "transient", "translate", "translate_regex",
        "translation", "treat", "trigger", "trim", "trim_array", "true",
        "truncate", "try_cast", "type", "uescape", "unbounded", "under", "undo",
        "union", "unique", "unknown", "unnest", "until", "update", "upper",
        "usage", "user", "using", "value", "value_of", "values", "var_pop",
        "var_samp", "varbinary", "varchar", "varying", "versioning", "view",
        "when", "whenever", "where", "while", "width_bucket", "window", "with",
        "within", "without", "work", "write", "year", "zone"]

    def __init__(self, cmd):
        self.cmd = cmd

    def get_command_completions(self, line):
        if ' ' not in line:
            cmd = line[1:]
            return (i for i in self.cmd.commands.keys() if i.startswith(cmd))
        parts = line.split(' ', 1)
        cmd = parts[0].lstrip('\\')
        cmd = self.cmd.commands.get(cmd, None)
        if isinstance(cmd, Command):
            return cmd.complete(self.cmd, parts[1])
        return []

    def get_completions(self, document, complete_event):
        if not self.cmd.should_autocomplete():
            return
        line = document.text
        word_before_cursor = document.get_word_before_cursor()
        if line.startswith('\\'):
            for w in self.get_command_completions(line):
                yield Completion(w, -len(word_before_cursor))
            return
        # start autocomplete on 3rd character for non-command keys
        if len(word_before_cursor) >= 3:
            for keyword in self.keywords:
                if keyword.startswith(word_before_cursor.lower()):
                    yield Completion(keyword.upper(), -len(word_before_cursor))

class CrashBuffer(Buffer):

    def __init__(self, *args, **kwargs):

        @Condition
        def is_multiline():
            doc = self.document
            if not doc.text:
                return False
            if doc.text.startswith('\\'):
                return False
            return not doc.text.rstrip().endswith(';')

        super(self.__class__, self).__init__(
            *args, is_multiline=is_multiline, **kwargs)


class Capitalizer:

    KEYWORD_RE = r'(?:"\w+)|(?:\'\w+)|\w+'

    def __init__(self, cmd):
        self.cmd = cmd
        self.last_changed = None

    def apply_capitalization(self, buffer):
        if not self.cmd.should_autocapitalize():
            return

        current_line = buffer.document.text

        if current_line.startswith('\\'):
            return

        cursor_position = buffer.document.cursor_position

        if self.last_changed and self.is_prefix(current_line[:cursor_position].lower(), self.last_changed.lower()):
            diff = len(self.last_changed) - len(current_line)
            current_line = self.last_changed + current_line[diff:]

        new_line = re.sub(self.KEYWORD_RE, self.keyword_replacer, current_line[:cursor_position])

        if new_line != buffer.document.text:
            buffer.delete_before_cursor(cursor_position)
            buffer.delete(len(new_line) - cursor_position)
            buffer.insert_text(new_line, overwrite=False, move_cursor=True, fire_event=False)
            self.last_changed = current_line[:cursor_position]

    def keyword_replacer(self, match):
        if match.group(0).lower() in SQLCompleter.keywords:
            return match.group(0).upper()
        else:
            return match.group(0)

    def is_prefix(self, string, prefix):
        return string.startswith(prefix) and string != prefix


def create_buffer(cmd, history_file):
    buffer = CrashBuffer(
        history=TruncatedFileHistory(history_file, max_length=MAX_HISTORY_LENGTH),
        accept_action=AcceptAction.RETURN_DOCUMENT,
        completer=SQLCompleter(cmd),
        on_text_insert=Capitalizer(cmd).apply_capitalization
    )
    buffer.complete_while_typing = lambda cli=None: cmd.should_autocomplete()
    return buffer


def _bind_keys(registry):

    def _is_start_of_multiline(cli):
        doc = cli.current_buffer.document
        word = doc.get_word_before_cursor()
        return doc.text and not word

    @registry.add_binding(Keys.Tab, filter=Condition(_is_start_of_multiline))
    def _(event):
        event.cli.current_buffer.insert_text('    ')

    def _bs_should_deindent(cli):
        doc = cli.current_buffer.document
        start_of_line_pos = doc.get_start_of_line_position()
        return (
            not doc.get_word_before_cursor() and
            start_of_line_pos != 0 and
            start_of_line_pos % 4 == 0
        )

    @registry.add_binding(Keys.Backspace, filter=Condition(_bs_should_deindent))
    def _(event):
        event.cli.current_buffer.delete_before_cursor(4)


def _get_toolbar_tokens(is_conn_available, username, active_servers):
        tokens = []
        if is_conn_available():
            active_servers = active_servers
            hosts = ', '.join(
                (n.replace('http://', '').replace('https://', '') for n in active_servers))
            tokens.extend([(Token.Toolbar.Status.Key, 'USER: '),
                           (Token.Toolbar.Status, '{}'.format(username)),
                           (Token.Toolbar.Status, ' | '),
                           (Token.Toolbar.Status.Key, 'HOSTS: '),
                           (Token.Toolbar.Status, hosts)])
        else:
            tokens.extend([(Token.Toolbar.Status, 'not connected')])
        return tokens


def loop(cmd, history_file):

    def session_toolbar(cli):
        return _get_toolbar_tokens(cmd.is_conn_available, cmd.username, cmd.connection.client.active_servers)

    key_binding_manager = KeyBindingManager(
        enable_search=True,
        enable_abort_and_exit_bindings=True,
        enable_system_bindings=True,
        enable_open_in_editor=True
    )
    _bind_keys(key_binding_manager.registry)
    layout = create_layout(
        message=u'cr> ',
        multiline=True,
        lexer=SqlLexer,
        extra_input_processors=[
            ConditionalProcessor(
                processor=HighlightMatchingBracketProcessor(chars='[](){}'),
                filter=HasFocus(DEFAULT_BUFFER) & ~IsDone())
        ],
        get_bottom_toolbar_tokens=session_toolbar
    )
    application = Application(
        layout=layout,
        buffer=create_buffer(cmd, history_file),
        style=PygmentsStyle.from_defaults(pygments_style_cls=CrateStyle),
        key_bindings_registry=key_binding_manager.registry,
        editing_mode=_get_editing_mode(),
        on_exit=AbortAction.RAISE_EXCEPTION,
        on_abort=AbortAction.RETRY,
    )
    eventloop = create_eventloop()
    output = create_output()
    cli = CommandLineInterface(
        application=application,
        eventloop=eventloop,
        output=output
    )

    def get_num_columns_override():
        return output.get_size().columns
    cmd.get_num_columns = get_num_columns_override

    while True:
        try:
            doc = cli.run(reset_current_buffer=True)
            if doc:
                cmd.process(doc.text)
        except KeyboardInterrupt:
            cmd.logger.warn("Query not cancelled. Run KILL <jobId> to cancel it")
        except EOFError:
            cmd.logger.warn(u'Bye!')
            return



# vim: set fileencodings=utf-8
# -*- coding: utf-8; -*-
# PYTHON_ARGCOMPLETE_OK
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


import os
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from functools import partial


class ConfigurationError(Exception):
    pass


class Configuration(object):
    """
    Model that reads default values for the CLI argument parser
    from a configuration file.
    """

    @classmethod
    def bwc_bool_transform_from(cls, x):
        """
        Read boolean values from old config files correctly
        and interpret 'True' and 'False' as correct booleans.
        """
        if x.lower() == 'true':
            return True
        elif x.lower() == 'false':
            return False
        return bool(int(x))

    def __init__(self, path):
        self.type_mapping = {
            str: partial(self._get_or_set,
                         transform_from=lambda x: str(x),
                         transform_to=lambda x: str(x)),
            int: partial(self._get_or_set,
                         transform_from=lambda x: int(x),
                         transform_to=lambda x: str(x)),
            bool: partial(self._get_or_set,
                          transform_from=Configuration.bwc_bool_transform_from,
                          transform_to=lambda x: str(int(x))),
            list: partial(self._get_or_set,
                          transform_from=lambda x: x.split('\n'),
                          transform_to=lambda x: '\n'.join(x)),
        }
        if not path.endswith('.cfg'):
            raise ConfigurationError('Path to configuration file needs to end with .cfg')
        self.path = path
        self.cfg = configparser.ConfigParser()
        self.read_and_create_if_necessary()
        self.add_crash_section_if_necessary()

    def read_and_create_if_necessary(self):
        dir = os.path.dirname(self.path)
        if dir and not os.path.exists(dir):
            os.makedirs(dir)
        if not os.path.exists(self.path):
            self.save()
        self.cfg.read(self.path)

    def add_crash_section_if_necessary(self):
        if 'crash' not in self.cfg.sections():
            self.cfg.add_section('crash')

    def get_or_set(self, key, default_value):
        option_type = type(default_value)
        if option_type in self.type_mapping:
            return self.type_mapping[option_type](key, default_value)
        return self._get_or_set(key, default_value)

    def _get_or_set(self, key,
                    default_value=None,
                    transform_from=lambda x: x,
                    transform_to=lambda x: x):
        assert 'crash' in self.cfg.sections()
        value = None
        try:
            value = self.cfg.get('crash', key)
        except configparser.NoOptionError:
            if default_value is not None:
                self.cfg.set('crash', key, transform_to(default_value))
        return default_value if value is None else transform_from(value)

    def save(self):
        with open(self.path, 'w') as fp:
            self.cfg.write(fp)




# vim: set fileencodings=utf-8
# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


import logging
from collections import namedtuple
from distutils.version import StrictVersion

from crate.client.exceptions import ConnectionError, ProgrammingError
from crate.client import connect

from .printer import ColorPrinter


Result = namedtuple('Result', ['rows', 'cols'])
SYSINFO_MIN_VERSION = StrictVersion("0.54.0")


class SysInfoCommand(object):

    CLUSTER_INFO = {
        'shards_query': """
select count(*) as number_of_shards, cast(sum(num_docs) as long) as number_of_records
from sys.shards
where "primary" = true
        """,
        'nodes_query': """
select count(*) as number_of_nodes
from sys.nodes
        """,
    }

    NODES_INFO = [ """ select name,
                          hostname,
                          version['number'] as crate_version,
                          round(heap['max'] / 1024.0 / 1024.0)
                                as total_heap_mb,
                          round((mem['free'] + mem['used']) / 1024.0 / 1024.0)
                                as total_memory_mb,
                          os_info['available_processors'] as cpus,
                          os['uptime'] /1000 as uptime_s,
                          format('%s - %s (%s)',
                                os_info['name'],
                                os_info['version'],
                                os_info['arch']) as os_info,
                          format('java version \"%s\" %s %s (build %s)',
                                os_info['jvm']['version'],
                                os_info['jvm']['vm_vendor'],
                                os_info['jvm']['vm_name'],
                                os_info['jvm']['vm_version']) as jvm_info
                          from sys.nodes
                          order by os['uptime'] desc """ ]

    def __init__(self, cmd):
        self.cmd = cmd

    def execute(self):
        """ print system and cluster info """
        if not self.cmd.is_conn_available():
            return
        if self.cmd.connection.lowest_server_version >= SYSINFO_MIN_VERSION:
            success, rows = self._sys_info()
            self.cmd.exit_code = self.cmd.exit_code or int(not success)
            if success:
                for result in rows:
                    self.cmd.pprint(result.rows, result.cols)
                self.cmd.logger.info(
                    "For debugging purposes you can send above listed information to support@crate.io")
        else:
            tmpl = 'Crate {version} does not support the cluster "sysinfo" command'
            self.cmd.logger.warn(tmpl
                                 .format(version=self.cmd.connection.lowest_server_version))

    def _sys_info(self):
        result = []
        success = self._cluster_info(result)
        success &= self._nodes_info(result)
        if success is False:
            result = []
        return (success, result)

    def _cluster_info(self, result):
        rows = []
        cols = []

        for query in SysInfoCommand.CLUSTER_INFO:
            success = self.cmd._execute(SysInfoCommand.CLUSTER_INFO[query])
            if success is False:
                return success
            rows.extend(self.cmd.cursor.fetchall()[0])
            cols.extend([c[0] for c in self.cmd.cursor.description])
        result.append(Result([rows], cols))
        return True

    def _nodes_info(self, result):
        success = self.cmd._execute(SysInfoCommand.NODES_INFO[0])
        if success:
            result.append(Result(self.cmd.cursor.fetchall(),
                                 [c[0] for c in self.cmd.cursor.description]))
        return success



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import re
from unittest import TestCase
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.token import Token
from .repl import SQLCompleter, Capitalizer, create_buffer, _get_toolbar_tokens
from .command import CrateCmd

class SQLCompleterTest(TestCase):

    def setUp(self):
        cmd = CrateCmd()
        self.completer = SQLCompleter(cmd)

    def test_get_builtin_command_completions(self):
        result = sorted(list(self.completer.get_command_completions('\\c')))
        self.assertEqual(result, ['c', 'check', 'connect'])

    def test_get_command_completions_format(self):
        result = list(self.completer.get_command_completions(r'\format dyn'))
        self.assertEqual(result, ['dynamic'])


class CrashBufferTest(TestCase):

    def test_create_buffer(self):
        cmd = CrateCmd()
        buffer = create_buffer(cmd, '/tmp/history')
        self.assertEquals(buffer.on_text_insert.fire(), None)


class AutoCapitalizeTest(TestCase):

    def setUp(self):
        cmd = CrateCmd()
        self.capitalizer = Capitalizer(cmd)

    def test_capitalize(self):
        buffer = Buffer()

        text = u'selec'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'selec', buffer.text)

        text = u'select'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'SELECT', buffer.text)

        text = u'CREATE TABLE "select'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'CREATE TABLE "select', buffer.text)

        text = u'CREATE TABLE \'select\''
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'CREATE TABLE \'select\'', buffer.text)

        text = u'create table test (x int)'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'CREATE TABLE test (x INT)', buffer.text)

        text = u'create table test (a boolean, b string, c integer)'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'CREATE TABLE test (a BOOLEAN, b STRING, c INTEGER)', buffer.text)

        text = u'create table test\n(a boolean, b string, c integer)'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'CREATE TABLE test\n(a BOOLEAN, b STRING, c INTEGER)', buffer.text)

        text = u'\select dynamic'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'\select dynamic', buffer.text)

    def test_undo_capitalize(self):
        buffer = Buffer()

        text = u'Selec'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'Selec', buffer.text)

        text = buffer.text + 't'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'SELECT', buffer.text)

        text = buffer.text + 'i'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'Selecti', buffer.text)

        text = buffer.text + 'on'
        buffer.set_document(Document(text, len(text)))
        self.capitalizer.apply_capitalization(buffer)
        self.assertEqual(u'Selection', buffer.text)

    def test_keyword_replacer(self):
        KEYWORD_RE = r'\w+'
        text = u'select'
        match = re.match(KEYWORD_RE, text)
        self.assertEqual(u'SELECT', self.capitalizer.keyword_replacer(match))

        text = u'definitelyNotAnSQLKeyword'
        match = re.match(KEYWORD_RE, text)
        self.assertEqual(u'definitelyNotAnSQLKeyword', self.capitalizer.keyword_replacer(match))

    def test_is_prefix(self):
        string = u'SELECT * FROM test'
        prefix = u'SELECT * FROM'
        self.assertTrue(self.capitalizer.is_prefix(string, prefix))

        prefix = u'SELECT testCol FROM'
        self.assertFalse(self.capitalizer.is_prefix(string, prefix))


class ToolbarTest(TestCase):

    def test_get_session_tokens(self):
        result = _get_toolbar_tokens(lambda: True, 'crate', ['http://host1:4200', 'https://host2:4200'])
        self.assertEqual(result, [(Token.Toolbar.Status.Key, 'USER: '),
                           (Token.Toolbar.Status, 'crate'),
                           (Token.Toolbar.Status, ' | '),
                           (Token.Toolbar.Status.Key, 'HOSTS: '),
                           (Token.Toolbar.Status, 'host1:4200, host2:4200')])

    def test_get_not_connected_tokens(self):
        result = _get_toolbar_tokens(lambda: False, 'crate', None)
        self.assertEqual(result, [(Token.Toolbar.Status, 'not connected')])



# vim: set fileencodings=utf-8
# -*- coding: utf-8; -*-
# PYTHON_ARGCOMPLETE_OK
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from prompt_toolkit.filters import IsDone, HasFocus, RendererHeightIsKnown, to_cli_filter, Condition
from prompt_toolkit.enums import DEFAULT_BUFFER, SEARCH_BUFFER
from prompt_toolkit.token import Token
from prompt_toolkit.layout import Window, HSplit, VSplit, Float
from prompt_toolkit.layout.containers import ConditionalContainer, FloatContainer
from prompt_toolkit.layout.dimension import LayoutDimension
from prompt_toolkit.layout.controls import TokenListControl, BufferControl, FillControl
from prompt_toolkit.layout.lexers import PygmentsLexer
from prompt_toolkit.layout.screen import Char
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.layout.processors import ConditionalProcessor, HighlightSearchProcessor
from prompt_toolkit.layout.toolbars import TokenListToolbar, SearchToolbar
from prompt_toolkit.layout.margins import PromptMargin, ConditionalMargin
from prompt_toolkit.layout.utils import token_list_width

""" Creates a custom `Layout` for the Crash input REPL

    This layout includes:
        * a bottom left-aligned session toolbar container
        * a bottom right-aligned side-bar container

    +-------------------------------------------+
    | cr> select 1;                             |
    |                                           |
    |                                           |
    +-------------------------------------------+
    | bottom_toolbar_tokens      sidebar_tokens |
    +-------------------------------------------+
"""
def create_layout(message='', lexer=None,
                  reserve_space_for_menu=8,
                  get_prompt_tokens=None,
                  get_bottom_toolbar_tokens=None,
                  extra_input_processors=None, multiline=False,
                  wrap_lines=True):

    # Create processors list.
    input_processors = [
        ConditionalProcessor(
            # Highlight the reverse-i-search buffer
            HighlightSearchProcessor(preview_search=True),
            HasFocus(SEARCH_BUFFER)),
    ]

    if extra_input_processors:
        input_processors.extend(extra_input_processors)

    lexer = PygmentsLexer(lexer, sync_from_start=True)
    multiline = to_cli_filter(multiline)

    sidebar_token = [
        (Token.Toolbar.Status.Key, "[ctrl+d]"),
        (Token.Toolbar.Status, " Exit")
    ]
    sidebar_width = token_list_width(sidebar_token)

    get_prompt_tokens = lambda _: [(Token.Prompt, message)]
    get_sidebar_tokens = lambda _: sidebar_token

    def get_height(cli):
        # If there is an autocompletion menu to be shown, make sure that our
        # layout has at least a minimal height in order to display it.
        if reserve_space_for_menu and not cli.is_done:
            buff = cli.current_buffer

            # Reserve the space, either when there are completions, or when
            # `complete_while_typing` is true and we expect completions very
            # soon.
            if buff.complete_while_typing() or buff.complete_state is not None:
                return LayoutDimension(min=reserve_space_for_menu)

        return LayoutDimension()

    # Create and return Container instance.
    return HSplit([
        VSplit([
            HSplit([
                # The main input, with completion menus floating on top of it.
                FloatContainer(
                    HSplit([
                        Window(
                            BufferControl(
                                input_processors=input_processors,
                                lexer=lexer,
                                # enable preview search for reverse-i-search
                                preview_search=True),
                            get_height=get_height,
                            wrap_lines=wrap_lines,
                            left_margins=[
                                # In multiline mode, use the window margin to display
                                # the prompt and continuation tokens.
                                ConditionalMargin(
                                    PromptMargin(get_prompt_tokens),
                                    filter=multiline
                                )
                            ],
                        ),
                    ]),
                    [
                        # Completion menu
                        Float(xcursor=True,
                            ycursor=True,
                            content=CompletionsMenu(
                                max_height=16,
                                scroll_offset=1,
                                extra_filter=HasFocus(DEFAULT_BUFFER))),
                    ]
                ),

                # reverse-i-search toolbar (ctrl+r)
                ConditionalContainer(SearchToolbar(), multiline),
            ])
        ]),
    ] + [
        VSplit([
            # Left-Aligned Session Toolbar
            ConditionalContainer(
                Window(
                    TokenListControl(get_bottom_toolbar_tokens),
                    height=LayoutDimension.exact(1)
                ),
                filter=~IsDone() & RendererHeightIsKnown()),

            # Right-Aligned Container
            ConditionalContainer(
                Window(
                    TokenListControl(get_sidebar_tokens),
                    height=LayoutDimension.exact(1),
                    width=LayoutDimension.exact(sidebar_width)
                ),
                filter=~IsDone() & RendererHeightIsKnown())
        ])
    ])



# -*- coding: utf-8; -*-
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from setuptools import setup, find_packages
import os
import io
import re


requirements = [
    'colorama',
    'Pygments>=2.0',
    'crate>=0.11.2',
    'appdirs>=1.2,<2.0',
    'prompt-toolkit>=1.0,<1.1'
]


def read(path):
    path = os.path.join(os.path.dirname(__file__), path)
    with io.open(path, 'r', encoding='utf-8') as f:
        return f.read()


long_description = (
    read('README.rst') + '\n' +
    read(os.path.join('src', 'crate', 'crash', 'usage.txt'))
)

versionf_content = read(os.path.join('src', 'crate', 'crash', '__init__.py'))
version_rex = r'^__version__ = [\'"]([^\'"]*)[\'"]$'
m = re.search(version_rex, versionf_content, re.M)
if m:
    version = m.group(1)
else:
    raise RuntimeError('Unable to find version string')

setup(
    name='crash',
    version=version,
    url='https://github.com/crate/crash',
    author='CRATE Technology GmbH',
    author_email='office@crate.io',
    package_dir={'': 'src'},
    description='The Crate Data Shell',
    long_description=long_description,
    platforms=['any'],
    license='Apache License 2.0',
    keywords='crate db data client shell',
    packages=find_packages('src'),
    namespace_packages=['crate'],
    entry_points={
        'console_scripts': [
            'crash = crate.crash.command:main',
        ]
    },
    extras_require=dict(
        test=['crate[test]',
              'zc.customdoctests',
              'zope.testing',
              ],
        argcompletion=['argcomplete']
    ),
    install_requires=requirements,
    package_data={'': ['*.txt']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Database'
    ],
)
